// Copyright © 2019 Weald Technology Trading
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package multicodec

// codec is the structure holding codec information
type codec struct {
	id   uint64
	tag  string
	name string
}

var codecs map[string]*codec
var reverseCodecs map[uint64]*codec

func init() {
	codecs = make(map[string]*codec)
	reverseCodecs = make(map[uint64]*codec)

	codecs["identity"] = &codec{id: 0x00, name: "identity", tag: "multihash"}
	reverseCodecs[0x00] = codecs["identity"]
	codecs["cidv1"] = &codec{id: 0x01, name: "cidv1", tag: "ipld"}
	reverseCodecs[0x01] = codecs["cidv1"]
	codecs["cidv2"] = &codec{id: 0x02, name: "cidv2", tag: "ipld"}
	reverseCodecs[0x02] = codecs["cidv2"]
	codecs["cidv3"] = &codec{id: 0x03, name: "cidv3", tag: "ipld"}
	reverseCodecs[0x03] = codecs["cidv3"]
	codecs["ip4"] = &codec{id: 0x04, name: "ip4", tag: "multiaddr"}
	reverseCodecs[0x04] = codecs["ip4"]
	codecs["tcp"] = &codec{id: 0x06, name: "tcp", tag: "multiaddr"}
	reverseCodecs[0x06] = codecs["tcp"]
	codecs["sha1"] = &codec{id: 0x11, name: "sha1", tag: "multihash"}
	reverseCodecs[0x11] = codecs["sha1"]
	codecs["sha2-256"] = &codec{id: 0x12, name: "sha2-256", tag: "multihash"}
	reverseCodecs[0x12] = codecs["sha2-256"]
	codecs["sha2-512"] = &codec{id: 0x13, name: "sha2-512", tag: "multihash"}
	reverseCodecs[0x13] = codecs["sha2-512"]
	codecs["sha3-512"] = &codec{id: 0x14, name: "sha3-512", tag: "multihash"}
	reverseCodecs[0x14] = codecs["sha3-512"]
	codecs["sha3-384"] = &codec{id: 0x15, name: "sha3-384", tag: "multihash"}
	reverseCodecs[0x15] = codecs["sha3-384"]
	codecs["sha3-256"] = &codec{id: 0x16, name: "sha3-256", tag: "multihash"}
	reverseCodecs[0x16] = codecs["sha3-256"]
	codecs["sha3-224"] = &codec{id: 0x17, name: "sha3-224", tag: "multihash"}
	reverseCodecs[0x17] = codecs["sha3-224"]
	codecs["shake-128"] = &codec{id: 0x18, name: "shake-128", tag: "multihash"}
	reverseCodecs[0x18] = codecs["shake-128"]
	codecs["shake-256"] = &codec{id: 0x19, name: "shake-256", tag: "multihash"}
	reverseCodecs[0x19] = codecs["shake-256"]
	codecs["keccak-224"] = &codec{id: 0x1a, name: "keccak-224", tag: "multihash"}
	reverseCodecs[0x1a] = codecs["keccak-224"]
	codecs["keccak-256"] = &codec{id: 0x1b, name: "keccak-256", tag: "multihash"}
	reverseCodecs[0x1b] = codecs["keccak-256"]
	codecs["keccak-384"] = &codec{id: 0x1c, name: "keccak-384", tag: "multihash"}
	reverseCodecs[0x1c] = codecs["keccak-384"]
	codecs["keccak-512"] = &codec{id: 0x1d, name: "keccak-512", tag: "multihash"}
	reverseCodecs[0x1d] = codecs["keccak-512"]
	codecs["blake3"] = &codec{id: 0x1e, name: "blake3", tag: "multihash"}
	reverseCodecs[0x1e] = codecs["blake3"]
	codecs["dccp"] = &codec{id: 0x21, name: "dccp", tag: "multiaddr"}
	reverseCodecs[0x21] = codecs["dccp"]
	codecs["murmur3-128"] = &codec{id: 0x22, name: "murmur3-128", tag: "multihash"}
	reverseCodecs[0x22] = codecs["murmur3-128"]
	codecs["murmur3-32"] = &codec{id: 0x23, name: "murmur3-32", tag: "multihash"}
	reverseCodecs[0x23] = codecs["murmur3-32"]
	codecs["ip6"] = &codec{id: 0x29, name: "ip6", tag: "multiaddr"}
	reverseCodecs[0x29] = codecs["ip6"]
	codecs["ip6zone"] = &codec{id: 0x2a, name: "ip6zone", tag: "multiaddr"}
	reverseCodecs[0x2a] = codecs["ip6zone"]
	codecs["path"] = &codec{id: 0x2f, name: "path", tag: "namespace"}
	reverseCodecs[0x2f] = codecs["path"]
	codecs["multicodec"] = &codec{id: 0x30, name: "multicodec", tag: "multiformat"}
	reverseCodecs[0x30] = codecs["multicodec"]
	codecs["multihash"] = &codec{id: 0x31, name: "multihash", tag: "multiformat"}
	reverseCodecs[0x31] = codecs["multihash"]
	codecs["multiaddr"] = &codec{id: 0x32, name: "multiaddr", tag: "multiformat"}
	reverseCodecs[0x32] = codecs["multiaddr"]
	codecs["multibase"] = &codec{id: 0x33, name: "multibase", tag: "multiformat"}
	reverseCodecs[0x33] = codecs["multibase"]
	codecs["dns"] = &codec{id: 0x35, name: "dns", tag: "multiaddr"}
	reverseCodecs[0x35] = codecs["dns"]
	codecs["dns4"] = &codec{id: 0x36, name: "dns4", tag: "multiaddr"}
	reverseCodecs[0x36] = codecs["dns4"]
	codecs["dns6"] = &codec{id: 0x37, name: "dns6", tag: "multiaddr"}
	reverseCodecs[0x37] = codecs["dns6"]
	codecs["dnsaddr"] = &codec{id: 0x38, name: "dnsaddr", tag: "multiaddr"}
	reverseCodecs[0x38] = codecs["dnsaddr"]
	codecs["protobuf"] = &codec{id: 0x50, name: "protobuf", tag: "serialization"}
	reverseCodecs[0x50] = codecs["protobuf"]
	codecs["cbor"] = &codec{id: 0x51, name: "cbor", tag: "serialization"}
	reverseCodecs[0x51] = codecs["cbor"]
	codecs["raw"] = &codec{id: 0x55, name: "raw", tag: "ipld"}
	reverseCodecs[0x55] = codecs["raw"]
	codecs["dbl-sha2-256"] = &codec{id: 0x56, name: "dbl-sha2-256", tag: "multihash"}
	reverseCodecs[0x56] = codecs["dbl-sha2-256"]
	codecs["rlp"] = &codec{id: 0x60, name: "rlp", tag: "serialization"}
	reverseCodecs[0x60] = codecs["rlp"]
	codecs["bencode"] = &codec{id: 0x63, name: "bencode", tag: "serialization"}
	reverseCodecs[0x63] = codecs["bencode"]
	codecs["dag-pb"] = &codec{id: 0x70, name: "dag-pb", tag: "ipld"}
	reverseCodecs[0x70] = codecs["dag-pb"]
	codecs["dag-cbor"] = &codec{id: 0x71, name: "dag-cbor", tag: "ipld"}
	reverseCodecs[0x71] = codecs["dag-cbor"]
	codecs["libp2p-key"] = &codec{id: 0x72, name: "libp2p-key", tag: "ipld"}
	reverseCodecs[0x72] = codecs["libp2p-key"]
	codecs["git-raw"] = &codec{id: 0x78, name: "git-raw", tag: "ipld"}
	reverseCodecs[0x78] = codecs["git-raw"]
	codecs["torrent-info"] = &codec{id: 0x7b, name: "torrent-info", tag: "ipld"}
	reverseCodecs[0x7b] = codecs["torrent-info"]
	codecs["torrent-file"] = &codec{id: 0x7c, name: "torrent-file", tag: "ipld"}
	reverseCodecs[0x7c] = codecs["torrent-file"]
	codecs["leofcoin-block"] = &codec{id: 0x81, name: "leofcoin-block", tag: "ipld"}
	reverseCodecs[0x81] = codecs["leofcoin-block"]
	codecs["leofcoin-tx"] = &codec{id: 0x82, name: "leofcoin-tx", tag: "ipld"}
	reverseCodecs[0x82] = codecs["leofcoin-tx"]
	codecs["leofcoin-pr"] = &codec{id: 0x83, name: "leofcoin-pr", tag: "ipld"}
	reverseCodecs[0x83] = codecs["leofcoin-pr"]
	codecs["sctp"] = &codec{id: 0x84, name: "sctp", tag: "multiaddr"}
	reverseCodecs[0x84] = codecs["sctp"]
	codecs["dag-jose"] = &codec{id: 0x85, name: "dag-jose", tag: "ipld"}
	reverseCodecs[0x85] = codecs["dag-jose"]
	codecs["dag-cose"] = &codec{id: 0x86, name: "dag-cose", tag: "ipld"}
	reverseCodecs[0x86] = codecs["dag-cose"]
	codecs["eth-block"] = &codec{id: 0x90, name: "eth-block", tag: "ipld"}
	reverseCodecs[0x90] = codecs["eth-block"]
	codecs["eth-block-list"] = &codec{id: 0x91, name: "eth-block-list", tag: "ipld"}
	reverseCodecs[0x91] = codecs["eth-block-list"]
	codecs["eth-tx-trie"] = &codec{id: 0x92, name: "eth-tx-trie", tag: "ipld"}
	reverseCodecs[0x92] = codecs["eth-tx-trie"]
	codecs["eth-tx"] = &codec{id: 0x93, name: "eth-tx", tag: "ipld"}
	reverseCodecs[0x93] = codecs["eth-tx"]
	codecs["eth-tx-receipt-trie"] = &codec{id: 0x94, name: "eth-tx-receipt-trie", tag: "ipld"}
	reverseCodecs[0x94] = codecs["eth-tx-receipt-trie"]
	codecs["eth-tx-receipt"] = &codec{id: 0x95, name: "eth-tx-receipt", tag: "ipld"}
	reverseCodecs[0x95] = codecs["eth-tx-receipt"]
	codecs["eth-state-trie"] = &codec{id: 0x96, name: "eth-state-trie", tag: "ipld"}
	reverseCodecs[0x96] = codecs["eth-state-trie"]
	codecs["eth-account-snapshot"] = &codec{id: 0x97, name: "eth-account-snapshot", tag: "ipld"}
	reverseCodecs[0x97] = codecs["eth-account-snapshot"]
	codecs["eth-storage-trie"] = &codec{id: 0x98, name: "eth-storage-trie", tag: "ipld"}
	reverseCodecs[0x98] = codecs["eth-storage-trie"]
	codecs["bitcoin-block"] = &codec{id: 0xb0, name: "bitcoin-block", tag: "ipld"}
	reverseCodecs[0xb0] = codecs["bitcoin-block"]
	codecs["bitcoin-tx"] = &codec{id: 0xb1, name: "bitcoin-tx", tag: "ipld"}
	reverseCodecs[0xb1] = codecs["bitcoin-tx"]
	codecs["bitcoin-witness-commitment"] = &codec{id: 0xb2, name: "bitcoin-witness-commitment", tag: "ipld"}
	reverseCodecs[0xb2] = codecs["bitcoin-witness-commitment"]
	codecs["zcash-block"] = &codec{id: 0xc0, name: "zcash-block", tag: "ipld"}
	reverseCodecs[0xc0] = codecs["zcash-block"]
	codecs["zcash-tx"] = &codec{id: 0xc1, name: "zcash-tx", tag: "ipld"}
	reverseCodecs[0xc1] = codecs["zcash-tx"]
	codecs["caip-50"] = &codec{id: 0xca, name: "caip-50", tag: "multiformat"}
	reverseCodecs[0xca] = codecs["caip-50"]
	codecs["streamid"] = &codec{id: 0xce, name: "streamid", tag: "namespace"}
	reverseCodecs[0xce] = codecs["streamid"]
	codecs["stellar-block"] = &codec{id: 0xd0, name: "stellar-block", tag: "ipld"}
	reverseCodecs[0xd0] = codecs["stellar-block"]
	codecs["stellar-tx"] = &codec{id: 0xd1, name: "stellar-tx", tag: "ipld"}
	reverseCodecs[0xd1] = codecs["stellar-tx"]
	codecs["md4"] = &codec{id: 0xd4, name: "md4", tag: "multihash"}
	reverseCodecs[0xd4] = codecs["md4"]
	codecs["md5"] = &codec{id: 0xd5, name: "md5", tag: "multihash"}
	reverseCodecs[0xd5] = codecs["md5"]
	codecs["bmt"] = &codec{id: 0xd6, name: "bmt", tag: "multihash"}
	reverseCodecs[0xd6] = codecs["bmt"]
	codecs["decred-block"] = &codec{id: 0xe0, name: "decred-block", tag: "ipld"}
	reverseCodecs[0xe0] = codecs["decred-block"]
	codecs["decred-tx"] = &codec{id: 0xe1, name: "decred-tx", tag: "ipld"}
	reverseCodecs[0xe1] = codecs["decred-tx"]
	codecs["ipld-ns"] = &codec{id: 0xe2, name: "ipld-ns", tag: "namespace"}
	reverseCodecs[0xe2] = codecs["ipld-ns"]
	codecs["ipfs-ns"] = &codec{id: 0xe3, name: "ipfs-ns", tag: "namespace"}
	reverseCodecs[0xe3] = codecs["ipfs-ns"]
	codecs["swarm-ns"] = &codec{id: 0xe4, name: "swarm-ns", tag: "namespace"}
	reverseCodecs[0xe4] = codecs["swarm-ns"]
	codecs["ipns-ns"] = &codec{id: 0xe5, name: "ipns-ns", tag: "namespace"}
	reverseCodecs[0xe5] = codecs["ipns-ns"]
	codecs["zeronet"] = &codec{id: 0xe6, name: "zeronet", tag: "namespace"}
	reverseCodecs[0xe6] = codecs["zeronet"]
	codecs["secp256k1-pub"] = &codec{id: 0xe7, name: "secp256k1-pub", tag: "key"}
	reverseCodecs[0xe7] = codecs["secp256k1-pub"]
	codecs["bls12_381-g1-pub"] = &codec{id: 0xea, name: "bls12_381-g1-pub", tag: "key"}
	reverseCodecs[0xea] = codecs["bls12_381-g1-pub"]
	codecs["bls12_381-g2-pub"] = &codec{id: 0xeb, name: "bls12_381-g2-pub", tag: "key"}
	reverseCodecs[0xeb] = codecs["bls12_381-g2-pub"]
	codecs["x25519-pub"] = &codec{id: 0xec, name: "x25519-pub", tag: "key"}
	reverseCodecs[0xec] = codecs["x25519-pub"]
	codecs["ed25519-pub"] = &codec{id: 0xed, name: "ed25519-pub", tag: "key"}
	reverseCodecs[0xed] = codecs["ed25519-pub"]
	codecs["bls12_381-g1g2-pub"] = &codec{id: 0xee, name: "bls12_381-g1g2-pub", tag: "key"}
	reverseCodecs[0xee] = codecs["bls12_381-g1g2-pub"]
	codecs["dash-block"] = &codec{id: 0xf0, name: "dash-block", tag: "ipld"}
	reverseCodecs[0xf0] = codecs["dash-block"]
	codecs["dash-tx"] = &codec{id: 0xf1, name: "dash-tx", tag: "ipld"}
	reverseCodecs[0xf1] = codecs["dash-tx"]
	codecs["swarm-manifest"] = &codec{id: 0xfa, name: "swarm-manifest", tag: "ipld"}
	reverseCodecs[0xfa] = codecs["swarm-manifest"]
	codecs["swarm-feed"] = &codec{id: 0xfb, name: "swarm-feed", tag: "ipld"}
	reverseCodecs[0xfb] = codecs["swarm-feed"]
	codecs["udp"] = &codec{id: 0x0111, name: "udp", tag: "multiaddr"}
	reverseCodecs[0x0111] = codecs["udp"]
	codecs["p2p-webrtc-star"] = &codec{id: 0x0113, name: "p2p-webrtc-star", tag: "multiaddr"}
	reverseCodecs[0x0113] = codecs["p2p-webrtc-star"]
	codecs["p2p-webrtc-direct"] = &codec{id: 0x0114, name: "p2p-webrtc-direct", tag: "multiaddr"}
	reverseCodecs[0x0114] = codecs["p2p-webrtc-direct"]
	codecs["p2p-stardust"] = &codec{id: 0x0115, name: "p2p-stardust", tag: "multiaddr"}
	reverseCodecs[0x0115] = codecs["p2p-stardust"]
	codecs["p2p-circuit"] = &codec{id: 0x0122, name: "p2p-circuit", tag: "multiaddr"}
	reverseCodecs[0x0122] = codecs["p2p-circuit"]
	codecs["dag-json"] = &codec{id: 0x0129, name: "dag-json", tag: "ipld"}
	reverseCodecs[0x0129] = codecs["dag-json"]
	codecs["udt"] = &codec{id: 0x012d, name: "udt", tag: "multiaddr"}
	reverseCodecs[0x012d] = codecs["udt"]
	codecs["utp"] = &codec{id: 0x012e, name: "utp", tag: "multiaddr"}
	reverseCodecs[0x012e] = codecs["utp"]
	codecs["unix"] = &codec{id: 0x0190, name: "unix", tag: "multiaddr"}
	reverseCodecs[0x0190] = codecs["unix"]
	codecs["thread"] = &codec{id: 0x0196, name: "thread", tag: "multiaddr"}
	reverseCodecs[0x0196] = codecs["thread"]
	codecs["p2p"] = &codec{id: 0x01a5, name: "p2p", tag: "multiaddr"}
	reverseCodecs[0x01a5] = codecs["p2p"]
	codecs["ipfs"] = &codec{id: 0x01a5, name: "ipfs", tag: "multiaddr"}
	codecs["https"] = &codec{id: 0x01bb, name: "https", tag: "multiaddr"}
	reverseCodecs[0x01bb] = codecs["https"]
	codecs["onion"] = &codec{id: 0x01bc, name: "onion", tag: "multiaddr"}
	reverseCodecs[0x01bc] = codecs["onion"]
	codecs["onion3"] = &codec{id: 0x01bd, name: "onion3", tag: "multiaddr"}
	reverseCodecs[0x01bd] = codecs["onion3"]
	codecs["garlic64"] = &codec{id: 0x01be, name: "garlic64", tag: "multiaddr"}
	reverseCodecs[0x01be] = codecs["garlic64"]
	codecs["garlic32"] = &codec{id: 0x01bf, name: "garlic32", tag: "multiaddr"}
	reverseCodecs[0x01bf] = codecs["garlic32"]
	codecs["tls"] = &codec{id: 0x01c0, name: "tls", tag: "multiaddr"}
	reverseCodecs[0x01c0] = codecs["tls"]
	codecs["quic"] = &codec{id: 0x01cc, name: "quic", tag: "multiaddr"}
	reverseCodecs[0x01cc] = codecs["quic"]
	codecs["ws"] = &codec{id: 0x01dd, name: "ws", tag: "multiaddr"}
	reverseCodecs[0x01dd] = codecs["ws"]
	codecs["wss"] = &codec{id: 0x01de, name: "wss", tag: "multiaddr"}
	reverseCodecs[0x01de] = codecs["wss"]
	codecs["p2p-websocket-star"] = &codec{id: 0x01df, name: "p2p-websocket-star", tag: "multiaddr"}
	reverseCodecs[0x01df] = codecs["p2p-websocket-star"]
	codecs["http"] = &codec{id: 0x01e0, name: "http", tag: "multiaddr"}
	reverseCodecs[0x01e0] = codecs["http"]
	codecs["json"] = &codec{id: 0x0200, name: "json", tag: "serialization"}
	reverseCodecs[0x0200] = codecs["json"]
	codecs["messagepack"] = &codec{id: 0x0201, name: "messagepack", tag: "serialization"}
	reverseCodecs[0x0201] = codecs["messagepack"]
	codecs["libp2p-peer-record"] = &codec{id: 0x0301, name: "libp2p-peer-record", tag: "libp2p"}
	reverseCodecs[0x0301] = codecs["libp2p-peer-record"]
	codecs["libp2p-relay-rsvp"] = &codec{id: 0x0302, name: "libp2p-relay-rsvp", tag: "libp2p"}
	reverseCodecs[0x0302] = codecs["libp2p-relay-rsvp"]
	codecs["sha2-256-trunc254-padded"] = &codec{id: 0x1012, name: "sha2-256-trunc254-padded", tag: "multihash"}
	reverseCodecs[0x1012] = codecs["sha2-256-trunc254-padded"]
	codecs["ripemd-128"] = &codec{id: 0x1052, name: "ripemd-128", tag: "multihash"}
	reverseCodecs[0x1052] = codecs["ripemd-128"]
	codecs["ripemd-160"] = &codec{id: 0x1053, name: "ripemd-160", tag: "multihash"}
	reverseCodecs[0x1053] = codecs["ripemd-160"]
	codecs["ripemd-256"] = &codec{id: 0x1054, name: "ripemd-256", tag: "multihash"}
	reverseCodecs[0x1054] = codecs["ripemd-256"]
	codecs["ripemd-320"] = &codec{id: 0x1055, name: "ripemd-320", tag: "multihash"}
	reverseCodecs[0x1055] = codecs["ripemd-320"]
	codecs["x11"] = &codec{id: 0x1100, name: "x11", tag: "multihash"}
	reverseCodecs[0x1100] = codecs["x11"]
	codecs["p256-pub"] = &codec{id: 0x1200, name: "p256-pub", tag: "key"}
	reverseCodecs[0x1200] = codecs["p256-pub"]
	codecs["p384-pub"] = &codec{id: 0x1201, name: "p384-pub", tag: "key"}
	reverseCodecs[0x1201] = codecs["p384-pub"]
	codecs["p521-pub"] = &codec{id: 0x1202, name: "p521-pub", tag: "key"}
	reverseCodecs[0x1202] = codecs["p521-pub"]
	codecs["ed448-pub"] = &codec{id: 0x1203, name: "ed448-pub", tag: "key"}
	reverseCodecs[0x1203] = codecs["ed448-pub"]
	codecs["x448-pub"] = &codec{id: 0x1204, name: "x448-pub", tag: "key"}
	reverseCodecs[0x1204] = codecs["x448-pub"]
	codecs["ed25519-priv"] = &codec{id: 0x1300, name: "ed25519-priv", tag: "key"}
	reverseCodecs[0x1300] = codecs["ed25519-priv"]
	codecs["secp256k1-priv"] = &codec{id: 0x1301, name: "secp256k1-priv", tag: "key"}
	reverseCodecs[0x1301] = codecs["secp256k1-priv"]
	codecs["x25519-priv"] = &codec{id: 0x1302, name: "x25519-priv", tag: "key"}
	reverseCodecs[0x1302] = codecs["x25519-priv"]
	codecs["kangarootwelve"] = &codec{id: 0x1d01, name: "kangarootwelve", tag: "multihash"}
	reverseCodecs[0x1d01] = codecs["kangarootwelve"]
	codecs["sm3-256"] = &codec{id: 0x534d, name: "sm3-256", tag: "multihash"}
	reverseCodecs[0x534d] = codecs["sm3-256"]
	codecs["blake2b-8"] = &codec{id: 0xb201, name: "blake2b-8", tag: "multihash"}
	reverseCodecs[0xb201] = codecs["blake2b-8"]
	codecs["blake2b-16"] = &codec{id: 0xb202, name: "blake2b-16", tag: "multihash"}
	reverseCodecs[0xb202] = codecs["blake2b-16"]
	codecs["blake2b-24"] = &codec{id: 0xb203, name: "blake2b-24", tag: "multihash"}
	reverseCodecs[0xb203] = codecs["blake2b-24"]
	codecs["blake2b-32"] = &codec{id: 0xb204, name: "blake2b-32", tag: "multihash"}
	reverseCodecs[0xb204] = codecs["blake2b-32"]
	codecs["blake2b-40"] = &codec{id: 0xb205, name: "blake2b-40", tag: "multihash"}
	reverseCodecs[0xb205] = codecs["blake2b-40"]
	codecs["blake2b-48"] = &codec{id: 0xb206, name: "blake2b-48", tag: "multihash"}
	reverseCodecs[0xb206] = codecs["blake2b-48"]
	codecs["blake2b-56"] = &codec{id: 0xb207, name: "blake2b-56", tag: "multihash"}
	reverseCodecs[0xb207] = codecs["blake2b-56"]
	codecs["blake2b-64"] = &codec{id: 0xb208, name: "blake2b-64", tag: "multihash"}
	reverseCodecs[0xb208] = codecs["blake2b-64"]
	codecs["blake2b-72"] = &codec{id: 0xb209, name: "blake2b-72", tag: "multihash"}
	reverseCodecs[0xb209] = codecs["blake2b-72"]
	codecs["blake2b-80"] = &codec{id: 0xb20a, name: "blake2b-80", tag: "multihash"}
	reverseCodecs[0xb20a] = codecs["blake2b-80"]
	codecs["blake2b-88"] = &codec{id: 0xb20b, name: "blake2b-88", tag: "multihash"}
	reverseCodecs[0xb20b] = codecs["blake2b-88"]
	codecs["blake2b-96"] = &codec{id: 0xb20c, name: "blake2b-96", tag: "multihash"}
	reverseCodecs[0xb20c] = codecs["blake2b-96"]
	codecs["blake2b-104"] = &codec{id: 0xb20d, name: "blake2b-104", tag: "multihash"}
	reverseCodecs[0xb20d] = codecs["blake2b-104"]
	codecs["blake2b-112"] = &codec{id: 0xb20e, name: "blake2b-112", tag: "multihash"}
	reverseCodecs[0xb20e] = codecs["blake2b-112"]
	codecs["blake2b-120"] = &codec{id: 0xb20f, name: "blake2b-120", tag: "multihash"}
	reverseCodecs[0xb20f] = codecs["blake2b-120"]
	codecs["blake2b-128"] = &codec{id: 0xb210, name: "blake2b-128", tag: "multihash"}
	reverseCodecs[0xb210] = codecs["blake2b-128"]
	codecs["blake2b-136"] = &codec{id: 0xb211, name: "blake2b-136", tag: "multihash"}
	reverseCodecs[0xb211] = codecs["blake2b-136"]
	codecs["blake2b-144"] = &codec{id: 0xb212, name: "blake2b-144", tag: "multihash"}
	reverseCodecs[0xb212] = codecs["blake2b-144"]
	codecs["blake2b-152"] = &codec{id: 0xb213, name: "blake2b-152", tag: "multihash"}
	reverseCodecs[0xb213] = codecs["blake2b-152"]
	codecs["blake2b-160"] = &codec{id: 0xb214, name: "blake2b-160", tag: "multihash"}
	reverseCodecs[0xb214] = codecs["blake2b-160"]
	codecs["blake2b-168"] = &codec{id: 0xb215, name: "blake2b-168", tag: "multihash"}
	reverseCodecs[0xb215] = codecs["blake2b-168"]
	codecs["blake2b-176"] = &codec{id: 0xb216, name: "blake2b-176", tag: "multihash"}
	reverseCodecs[0xb216] = codecs["blake2b-176"]
	codecs["blake2b-184"] = &codec{id: 0xb217, name: "blake2b-184", tag: "multihash"}
	reverseCodecs[0xb217] = codecs["blake2b-184"]
	codecs["blake2b-192"] = &codec{id: 0xb218, name: "blake2b-192", tag: "multihash"}
	reverseCodecs[0xb218] = codecs["blake2b-192"]
	codecs["blake2b-200"] = &codec{id: 0xb219, name: "blake2b-200", tag: "multihash"}
	reverseCodecs[0xb219] = codecs["blake2b-200"]
	codecs["blake2b-208"] = &codec{id: 0xb21a, name: "blake2b-208", tag: "multihash"}
	reverseCodecs[0xb21a] = codecs["blake2b-208"]
	codecs["blake2b-216"] = &codec{id: 0xb21b, name: "blake2b-216", tag: "multihash"}
	reverseCodecs[0xb21b] = codecs["blake2b-216"]
	codecs["blake2b-224"] = &codec{id: 0xb21c, name: "blake2b-224", tag: "multihash"}
	reverseCodecs[0xb21c] = codecs["blake2b-224"]
	codecs["blake2b-232"] = &codec{id: 0xb21d, name: "blake2b-232", tag: "multihash"}
	reverseCodecs[0xb21d] = codecs["blake2b-232"]
	codecs["blake2b-240"] = &codec{id: 0xb21e, name: "blake2b-240", tag: "multihash"}
	reverseCodecs[0xb21e] = codecs["blake2b-240"]
	codecs["blake2b-248"] = &codec{id: 0xb21f, name: "blake2b-248", tag: "multihash"}
	reverseCodecs[0xb21f] = codecs["blake2b-248"]
	codecs["blake2b-256"] = &codec{id: 0xb220, name: "blake2b-256", tag: "multihash"}
	reverseCodecs[0xb220] = codecs["blake2b-256"]
	codecs["blake2b-264"] = &codec{id: 0xb221, name: "blake2b-264", tag: "multihash"}
	reverseCodecs[0xb221] = codecs["blake2b-264"]
	codecs["blake2b-272"] = &codec{id: 0xb222, name: "blake2b-272", tag: "multihash"}
	reverseCodecs[0xb222] = codecs["blake2b-272"]
	codecs["blake2b-280"] = &codec{id: 0xb223, name: "blake2b-280", tag: "multihash"}
	reverseCodecs[0xb223] = codecs["blake2b-280"]
	codecs["blake2b-288"] = &codec{id: 0xb224, name: "blake2b-288", tag: "multihash"}
	reverseCodecs[0xb224] = codecs["blake2b-288"]
	codecs["blake2b-296"] = &codec{id: 0xb225, name: "blake2b-296", tag: "multihash"}
	reverseCodecs[0xb225] = codecs["blake2b-296"]
	codecs["blake2b-304"] = &codec{id: 0xb226, name: "blake2b-304", tag: "multihash"}
	reverseCodecs[0xb226] = codecs["blake2b-304"]
	codecs["blake2b-312"] = &codec{id: 0xb227, name: "blake2b-312", tag: "multihash"}
	reverseCodecs[0xb227] = codecs["blake2b-312"]
	codecs["blake2b-320"] = &codec{id: 0xb228, name: "blake2b-320", tag: "multihash"}
	reverseCodecs[0xb228] = codecs["blake2b-320"]
	codecs["blake2b-328"] = &codec{id: 0xb229, name: "blake2b-328", tag: "multihash"}
	reverseCodecs[0xb229] = codecs["blake2b-328"]
	codecs["blake2b-336"] = &codec{id: 0xb22a, name: "blake2b-336", tag: "multihash"}
	reverseCodecs[0xb22a] = codecs["blake2b-336"]
	codecs["blake2b-344"] = &codec{id: 0xb22b, name: "blake2b-344", tag: "multihash"}
	reverseCodecs[0xb22b] = codecs["blake2b-344"]
	codecs["blake2b-352"] = &codec{id: 0xb22c, name: "blake2b-352", tag: "multihash"}
	reverseCodecs[0xb22c] = codecs["blake2b-352"]
	codecs["blake2b-360"] = &codec{id: 0xb22d, name: "blake2b-360", tag: "multihash"}
	reverseCodecs[0xb22d] = codecs["blake2b-360"]
	codecs["blake2b-368"] = &codec{id: 0xb22e, name: "blake2b-368", tag: "multihash"}
	reverseCodecs[0xb22e] = codecs["blake2b-368"]
	codecs["blake2b-376"] = &codec{id: 0xb22f, name: "blake2b-376", tag: "multihash"}
	reverseCodecs[0xb22f] = codecs["blake2b-376"]
	codecs["blake2b-384"] = &codec{id: 0xb230, name: "blake2b-384", tag: "multihash"}
	reverseCodecs[0xb230] = codecs["blake2b-384"]
	codecs["blake2b-392"] = &codec{id: 0xb231, name: "blake2b-392", tag: "multihash"}
	reverseCodecs[0xb231] = codecs["blake2b-392"]
	codecs["blake2b-400"] = &codec{id: 0xb232, name: "blake2b-400", tag: "multihash"}
	reverseCodecs[0xb232] = codecs["blake2b-400"]
	codecs["blake2b-408"] = &codec{id: 0xb233, name: "blake2b-408", tag: "multihash"}
	reverseCodecs[0xb233] = codecs["blake2b-408"]
	codecs["blake2b-416"] = &codec{id: 0xb234, name: "blake2b-416", tag: "multihash"}
	reverseCodecs[0xb234] = codecs["blake2b-416"]
	codecs["blake2b-424"] = &codec{id: 0xb235, name: "blake2b-424", tag: "multihash"}
	reverseCodecs[0xb235] = codecs["blake2b-424"]
	codecs["blake2b-432"] = &codec{id: 0xb236, name: "blake2b-432", tag: "multihash"}
	reverseCodecs[0xb236] = codecs["blake2b-432"]
	codecs["blake2b-440"] = &codec{id: 0xb237, name: "blake2b-440", tag: "multihash"}
	reverseCodecs[0xb237] = codecs["blake2b-440"]
	codecs["blake2b-448"] = &codec{id: 0xb238, name: "blake2b-448", tag: "multihash"}
	reverseCodecs[0xb238] = codecs["blake2b-448"]
	codecs["blake2b-456"] = &codec{id: 0xb239, name: "blake2b-456", tag: "multihash"}
	reverseCodecs[0xb239] = codecs["blake2b-456"]
	codecs["blake2b-464"] = &codec{id: 0xb23a, name: "blake2b-464", tag: "multihash"}
	reverseCodecs[0xb23a] = codecs["blake2b-464"]
	codecs["blake2b-472"] = &codec{id: 0xb23b, name: "blake2b-472", tag: "multihash"}
	reverseCodecs[0xb23b] = codecs["blake2b-472"]
	codecs["blake2b-480"] = &codec{id: 0xb23c, name: "blake2b-480", tag: "multihash"}
	reverseCodecs[0xb23c] = codecs["blake2b-480"]
	codecs["blake2b-488"] = &codec{id: 0xb23d, name: "blake2b-488", tag: "multihash"}
	reverseCodecs[0xb23d] = codecs["blake2b-488"]
	codecs["blake2b-496"] = &codec{id: 0xb23e, name: "blake2b-496", tag: "multihash"}
	reverseCodecs[0xb23e] = codecs["blake2b-496"]
	codecs["blake2b-504"] = &codec{id: 0xb23f, name: "blake2b-504", tag: "multihash"}
	reverseCodecs[0xb23f] = codecs["blake2b-504"]
	codecs["blake2b-512"] = &codec{id: 0xb240, name: "blake2b-512", tag: "multihash"}
	reverseCodecs[0xb240] = codecs["blake2b-512"]
	codecs["blake2s-8"] = &codec{id: 0xb241, name: "blake2s-8", tag: "multihash"}
	reverseCodecs[0xb241] = codecs["blake2s-8"]
	codecs["blake2s-16"] = &codec{id: 0xb242, name: "blake2s-16", tag: "multihash"}
	reverseCodecs[0xb242] = codecs["blake2s-16"]
	codecs["blake2s-24"] = &codec{id: 0xb243, name: "blake2s-24", tag: "multihash"}
	reverseCodecs[0xb243] = codecs["blake2s-24"]
	codecs["blake2s-32"] = &codec{id: 0xb244, name: "blake2s-32", tag: "multihash"}
	reverseCodecs[0xb244] = codecs["blake2s-32"]
	codecs["blake2s-40"] = &codec{id: 0xb245, name: "blake2s-40", tag: "multihash"}
	reverseCodecs[0xb245] = codecs["blake2s-40"]
	codecs["blake2s-48"] = &codec{id: 0xb246, name: "blake2s-48", tag: "multihash"}
	reverseCodecs[0xb246] = codecs["blake2s-48"]
	codecs["blake2s-56"] = &codec{id: 0xb247, name: "blake2s-56", tag: "multihash"}
	reverseCodecs[0xb247] = codecs["blake2s-56"]
	codecs["blake2s-64"] = &codec{id: 0xb248, name: "blake2s-64", tag: "multihash"}
	reverseCodecs[0xb248] = codecs["blake2s-64"]
	codecs["blake2s-72"] = &codec{id: 0xb249, name: "blake2s-72", tag: "multihash"}
	reverseCodecs[0xb249] = codecs["blake2s-72"]
	codecs["blake2s-80"] = &codec{id: 0xb24a, name: "blake2s-80", tag: "multihash"}
	reverseCodecs[0xb24a] = codecs["blake2s-80"]
	codecs["blake2s-88"] = &codec{id: 0xb24b, name: "blake2s-88", tag: "multihash"}
	reverseCodecs[0xb24b] = codecs["blake2s-88"]
	codecs["blake2s-96"] = &codec{id: 0xb24c, name: "blake2s-96", tag: "multihash"}
	reverseCodecs[0xb24c] = codecs["blake2s-96"]
	codecs["blake2s-104"] = &codec{id: 0xb24d, name: "blake2s-104", tag: "multihash"}
	reverseCodecs[0xb24d] = codecs["blake2s-104"]
	codecs["blake2s-112"] = &codec{id: 0xb24e, name: "blake2s-112", tag: "multihash"}
	reverseCodecs[0xb24e] = codecs["blake2s-112"]
	codecs["blake2s-120"] = &codec{id: 0xb24f, name: "blake2s-120", tag: "multihash"}
	reverseCodecs[0xb24f] = codecs["blake2s-120"]
	codecs["blake2s-128"] = &codec{id: 0xb250, name: "blake2s-128", tag: "multihash"}
	reverseCodecs[0xb250] = codecs["blake2s-128"]
	codecs["blake2s-136"] = &codec{id: 0xb251, name: "blake2s-136", tag: "multihash"}
	reverseCodecs[0xb251] = codecs["blake2s-136"]
	codecs["blake2s-144"] = &codec{id: 0xb252, name: "blake2s-144", tag: "multihash"}
	reverseCodecs[0xb252] = codecs["blake2s-144"]
	codecs["blake2s-152"] = &codec{id: 0xb253, name: "blake2s-152", tag: "multihash"}
	reverseCodecs[0xb253] = codecs["blake2s-152"]
	codecs["blake2s-160"] = &codec{id: 0xb254, name: "blake2s-160", tag: "multihash"}
	reverseCodecs[0xb254] = codecs["blake2s-160"]
	codecs["blake2s-168"] = &codec{id: 0xb255, name: "blake2s-168", tag: "multihash"}
	reverseCodecs[0xb255] = codecs["blake2s-168"]
	codecs["blake2s-176"] = &codec{id: 0xb256, name: "blake2s-176", tag: "multihash"}
	reverseCodecs[0xb256] = codecs["blake2s-176"]
	codecs["blake2s-184"] = &codec{id: 0xb257, name: "blake2s-184", tag: "multihash"}
	reverseCodecs[0xb257] = codecs["blake2s-184"]
	codecs["blake2s-192"] = &codec{id: 0xb258, name: "blake2s-192", tag: "multihash"}
	reverseCodecs[0xb258] = codecs["blake2s-192"]
	codecs["blake2s-200"] = &codec{id: 0xb259, name: "blake2s-200", tag: "multihash"}
	reverseCodecs[0xb259] = codecs["blake2s-200"]
	codecs["blake2s-208"] = &codec{id: 0xb25a, name: "blake2s-208", tag: "multihash"}
	reverseCodecs[0xb25a] = codecs["blake2s-208"]
	codecs["blake2s-216"] = &codec{id: 0xb25b, name: "blake2s-216", tag: "multihash"}
	reverseCodecs[0xb25b] = codecs["blake2s-216"]
	codecs["blake2s-224"] = &codec{id: 0xb25c, name: "blake2s-224", tag: "multihash"}
	reverseCodecs[0xb25c] = codecs["blake2s-224"]
	codecs["blake2s-232"] = &codec{id: 0xb25d, name: "blake2s-232", tag: "multihash"}
	reverseCodecs[0xb25d] = codecs["blake2s-232"]
	codecs["blake2s-240"] = &codec{id: 0xb25e, name: "blake2s-240", tag: "multihash"}
	reverseCodecs[0xb25e] = codecs["blake2s-240"]
	codecs["blake2s-248"] = &codec{id: 0xb25f, name: "blake2s-248", tag: "multihash"}
	reverseCodecs[0xb25f] = codecs["blake2s-248"]
	codecs["blake2s-256"] = &codec{id: 0xb260, name: "blake2s-256", tag: "multihash"}
	reverseCodecs[0xb260] = codecs["blake2s-256"]
	codecs["skein256-8"] = &codec{id: 0xb301, name: "skein256-8", tag: "multihash"}
	reverseCodecs[0xb301] = codecs["skein256-8"]
	codecs["skein256-16"] = &codec{id: 0xb302, name: "skein256-16", tag: "multihash"}
	reverseCodecs[0xb302] = codecs["skein256-16"]
	codecs["skein256-24"] = &codec{id: 0xb303, name: "skein256-24", tag: "multihash"}
	reverseCodecs[0xb303] = codecs["skein256-24"]
	codecs["skein256-32"] = &codec{id: 0xb304, name: "skein256-32", tag: "multihash"}
	reverseCodecs[0xb304] = codecs["skein256-32"]
	codecs["skein256-40"] = &codec{id: 0xb305, name: "skein256-40", tag: "multihash"}
	reverseCodecs[0xb305] = codecs["skein256-40"]
	codecs["skein256-48"] = &codec{id: 0xb306, name: "skein256-48", tag: "multihash"}
	reverseCodecs[0xb306] = codecs["skein256-48"]
	codecs["skein256-56"] = &codec{id: 0xb307, name: "skein256-56", tag: "multihash"}
	reverseCodecs[0xb307] = codecs["skein256-56"]
	codecs["skein256-64"] = &codec{id: 0xb308, name: "skein256-64", tag: "multihash"}
	reverseCodecs[0xb308] = codecs["skein256-64"]
	codecs["skein256-72"] = &codec{id: 0xb309, name: "skein256-72", tag: "multihash"}
	reverseCodecs[0xb309] = codecs["skein256-72"]
	codecs["skein256-80"] = &codec{id: 0xb30a, name: "skein256-80", tag: "multihash"}
	reverseCodecs[0xb30a] = codecs["skein256-80"]
	codecs["skein256-88"] = &codec{id: 0xb30b, name: "skein256-88", tag: "multihash"}
	reverseCodecs[0xb30b] = codecs["skein256-88"]
	codecs["skein256-96"] = &codec{id: 0xb30c, name: "skein256-96", tag: "multihash"}
	reverseCodecs[0xb30c] = codecs["skein256-96"]
	codecs["skein256-104"] = &codec{id: 0xb30d, name: "skein256-104", tag: "multihash"}
	reverseCodecs[0xb30d] = codecs["skein256-104"]
	codecs["skein256-112"] = &codec{id: 0xb30e, name: "skein256-112", tag: "multihash"}
	reverseCodecs[0xb30e] = codecs["skein256-112"]
	codecs["skein256-120"] = &codec{id: 0xb30f, name: "skein256-120", tag: "multihash"}
	reverseCodecs[0xb30f] = codecs["skein256-120"]
	codecs["skein256-128"] = &codec{id: 0xb310, name: "skein256-128", tag: "multihash"}
	reverseCodecs[0xb310] = codecs["skein256-128"]
	codecs["skein256-136"] = &codec{id: 0xb311, name: "skein256-136", tag: "multihash"}
	reverseCodecs[0xb311] = codecs["skein256-136"]
	codecs["skein256-144"] = &codec{id: 0xb312, name: "skein256-144", tag: "multihash"}
	reverseCodecs[0xb312] = codecs["skein256-144"]
	codecs["skein256-152"] = &codec{id: 0xb313, name: "skein256-152", tag: "multihash"}
	reverseCodecs[0xb313] = codecs["skein256-152"]
	codecs["skein256-160"] = &codec{id: 0xb314, name: "skein256-160", tag: "multihash"}
	reverseCodecs[0xb314] = codecs["skein256-160"]
	codecs["skein256-168"] = &codec{id: 0xb315, name: "skein256-168", tag: "multihash"}
	reverseCodecs[0xb315] = codecs["skein256-168"]
	codecs["skein256-176"] = &codec{id: 0xb316, name: "skein256-176", tag: "multihash"}
	reverseCodecs[0xb316] = codecs["skein256-176"]
	codecs["skein256-184"] = &codec{id: 0xb317, name: "skein256-184", tag: "multihash"}
	reverseCodecs[0xb317] = codecs["skein256-184"]
	codecs["skein256-192"] = &codec{id: 0xb318, name: "skein256-192", tag: "multihash"}
	reverseCodecs[0xb318] = codecs["skein256-192"]
	codecs["skein256-200"] = &codec{id: 0xb319, name: "skein256-200", tag: "multihash"}
	reverseCodecs[0xb319] = codecs["skein256-200"]
	codecs["skein256-208"] = &codec{id: 0xb31a, name: "skein256-208", tag: "multihash"}
	reverseCodecs[0xb31a] = codecs["skein256-208"]
	codecs["skein256-216"] = &codec{id: 0xb31b, name: "skein256-216", tag: "multihash"}
	reverseCodecs[0xb31b] = codecs["skein256-216"]
	codecs["skein256-224"] = &codec{id: 0xb31c, name: "skein256-224", tag: "multihash"}
	reverseCodecs[0xb31c] = codecs["skein256-224"]
	codecs["skein256-232"] = &codec{id: 0xb31d, name: "skein256-232", tag: "multihash"}
	reverseCodecs[0xb31d] = codecs["skein256-232"]
	codecs["skein256-240"] = &codec{id: 0xb31e, name: "skein256-240", tag: "multihash"}
	reverseCodecs[0xb31e] = codecs["skein256-240"]
	codecs["skein256-248"] = &codec{id: 0xb31f, name: "skein256-248", tag: "multihash"}
	reverseCodecs[0xb31f] = codecs["skein256-248"]
	codecs["skein256-256"] = &codec{id: 0xb320, name: "skein256-256", tag: "multihash"}
	reverseCodecs[0xb320] = codecs["skein256-256"]
	codecs["skein512-8"] = &codec{id: 0xb321, name: "skein512-8", tag: "multihash"}
	reverseCodecs[0xb321] = codecs["skein512-8"]
	codecs["skein512-16"] = &codec{id: 0xb322, name: "skein512-16", tag: "multihash"}
	reverseCodecs[0xb322] = codecs["skein512-16"]
	codecs["skein512-24"] = &codec{id: 0xb323, name: "skein512-24", tag: "multihash"}
	reverseCodecs[0xb323] = codecs["skein512-24"]
	codecs["skein512-32"] = &codec{id: 0xb324, name: "skein512-32", tag: "multihash"}
	reverseCodecs[0xb324] = codecs["skein512-32"]
	codecs["skein512-40"] = &codec{id: 0xb325, name: "skein512-40", tag: "multihash"}
	reverseCodecs[0xb325] = codecs["skein512-40"]
	codecs["skein512-48"] = &codec{id: 0xb326, name: "skein512-48", tag: "multihash"}
	reverseCodecs[0xb326] = codecs["skein512-48"]
	codecs["skein512-56"] = &codec{id: 0xb327, name: "skein512-56", tag: "multihash"}
	reverseCodecs[0xb327] = codecs["skein512-56"]
	codecs["skein512-64"] = &codec{id: 0xb328, name: "skein512-64", tag: "multihash"}
	reverseCodecs[0xb328] = codecs["skein512-64"]
	codecs["skein512-72"] = &codec{id: 0xb329, name: "skein512-72", tag: "multihash"}
	reverseCodecs[0xb329] = codecs["skein512-72"]
	codecs["skein512-80"] = &codec{id: 0xb32a, name: "skein512-80", tag: "multihash"}
	reverseCodecs[0xb32a] = codecs["skein512-80"]
	codecs["skein512-88"] = &codec{id: 0xb32b, name: "skein512-88", tag: "multihash"}
	reverseCodecs[0xb32b] = codecs["skein512-88"]
	codecs["skein512-96"] = &codec{id: 0xb32c, name: "skein512-96", tag: "multihash"}
	reverseCodecs[0xb32c] = codecs["skein512-96"]
	codecs["skein512-104"] = &codec{id: 0xb32d, name: "skein512-104", tag: "multihash"}
	reverseCodecs[0xb32d] = codecs["skein512-104"]
	codecs["skein512-112"] = &codec{id: 0xb32e, name: "skein512-112", tag: "multihash"}
	reverseCodecs[0xb32e] = codecs["skein512-112"]
	codecs["skein512-120"] = &codec{id: 0xb32f, name: "skein512-120", tag: "multihash"}
	reverseCodecs[0xb32f] = codecs["skein512-120"]
	codecs["skein512-128"] = &codec{id: 0xb330, name: "skein512-128", tag: "multihash"}
	reverseCodecs[0xb330] = codecs["skein512-128"]
	codecs["skein512-136"] = &codec{id: 0xb331, name: "skein512-136", tag: "multihash"}
	reverseCodecs[0xb331] = codecs["skein512-136"]
	codecs["skein512-144"] = &codec{id: 0xb332, name: "skein512-144", tag: "multihash"}
	reverseCodecs[0xb332] = codecs["skein512-144"]
	codecs["skein512-152"] = &codec{id: 0xb333, name: "skein512-152", tag: "multihash"}
	reverseCodecs[0xb333] = codecs["skein512-152"]
	codecs["skein512-160"] = &codec{id: 0xb334, name: "skein512-160", tag: "multihash"}
	reverseCodecs[0xb334] = codecs["skein512-160"]
	codecs["skein512-168"] = &codec{id: 0xb335, name: "skein512-168", tag: "multihash"}
	reverseCodecs[0xb335] = codecs["skein512-168"]
	codecs["skein512-176"] = &codec{id: 0xb336, name: "skein512-176", tag: "multihash"}
	reverseCodecs[0xb336] = codecs["skein512-176"]
	codecs["skein512-184"] = &codec{id: 0xb337, name: "skein512-184", tag: "multihash"}
	reverseCodecs[0xb337] = codecs["skein512-184"]
	codecs["skein512-192"] = &codec{id: 0xb338, name: "skein512-192", tag: "multihash"}
	reverseCodecs[0xb338] = codecs["skein512-192"]
	codecs["skein512-200"] = &codec{id: 0xb339, name: "skein512-200", tag: "multihash"}
	reverseCodecs[0xb339] = codecs["skein512-200"]
	codecs["skein512-208"] = &codec{id: 0xb33a, name: "skein512-208", tag: "multihash"}
	reverseCodecs[0xb33a] = codecs["skein512-208"]
	codecs["skein512-216"] = &codec{id: 0xb33b, name: "skein512-216", tag: "multihash"}
	reverseCodecs[0xb33b] = codecs["skein512-216"]
	codecs["skein512-224"] = &codec{id: 0xb33c, name: "skein512-224", tag: "multihash"}
	reverseCodecs[0xb33c] = codecs["skein512-224"]
	codecs["skein512-232"] = &codec{id: 0xb33d, name: "skein512-232", tag: "multihash"}
	reverseCodecs[0xb33d] = codecs["skein512-232"]
	codecs["skein512-240"] = &codec{id: 0xb33e, name: "skein512-240", tag: "multihash"}
	reverseCodecs[0xb33e] = codecs["skein512-240"]
	codecs["skein512-248"] = &codec{id: 0xb33f, name: "skein512-248", tag: "multihash"}
	reverseCodecs[0xb33f] = codecs["skein512-248"]
	codecs["skein512-256"] = &codec{id: 0xb340, name: "skein512-256", tag: "multihash"}
	reverseCodecs[0xb340] = codecs["skein512-256"]
	codecs["skein512-264"] = &codec{id: 0xb341, name: "skein512-264", tag: "multihash"}
	reverseCodecs[0xb341] = codecs["skein512-264"]
	codecs["skein512-272"] = &codec{id: 0xb342, name: "skein512-272", tag: "multihash"}
	reverseCodecs[0xb342] = codecs["skein512-272"]
	codecs["skein512-280"] = &codec{id: 0xb343, name: "skein512-280", tag: "multihash"}
	reverseCodecs[0xb343] = codecs["skein512-280"]
	codecs["skein512-288"] = &codec{id: 0xb344, name: "skein512-288", tag: "multihash"}
	reverseCodecs[0xb344] = codecs["skein512-288"]
	codecs["skein512-296"] = &codec{id: 0xb345, name: "skein512-296", tag: "multihash"}
	reverseCodecs[0xb345] = codecs["skein512-296"]
	codecs["skein512-304"] = &codec{id: 0xb346, name: "skein512-304", tag: "multihash"}
	reverseCodecs[0xb346] = codecs["skein512-304"]
	codecs["skein512-312"] = &codec{id: 0xb347, name: "skein512-312", tag: "multihash"}
	reverseCodecs[0xb347] = codecs["skein512-312"]
	codecs["skein512-320"] = &codec{id: 0xb348, name: "skein512-320", tag: "multihash"}
	reverseCodecs[0xb348] = codecs["skein512-320"]
	codecs["skein512-328"] = &codec{id: 0xb349, name: "skein512-328", tag: "multihash"}
	reverseCodecs[0xb349] = codecs["skein512-328"]
	codecs["skein512-336"] = &codec{id: 0xb34a, name: "skein512-336", tag: "multihash"}
	reverseCodecs[0xb34a] = codecs["skein512-336"]
	codecs["skein512-344"] = &codec{id: 0xb34b, name: "skein512-344", tag: "multihash"}
	reverseCodecs[0xb34b] = codecs["skein512-344"]
	codecs["skein512-352"] = &codec{id: 0xb34c, name: "skein512-352", tag: "multihash"}
	reverseCodecs[0xb34c] = codecs["skein512-352"]
	codecs["skein512-360"] = &codec{id: 0xb34d, name: "skein512-360", tag: "multihash"}
	reverseCodecs[0xb34d] = codecs["skein512-360"]
	codecs["skein512-368"] = &codec{id: 0xb34e, name: "skein512-368", tag: "multihash"}
	reverseCodecs[0xb34e] = codecs["skein512-368"]
	codecs["skein512-376"] = &codec{id: 0xb34f, name: "skein512-376", tag: "multihash"}
	reverseCodecs[0xb34f] = codecs["skein512-376"]
	codecs["skein512-384"] = &codec{id: 0xb350, name: "skein512-384", tag: "multihash"}
	reverseCodecs[0xb350] = codecs["skein512-384"]
	codecs["skein512-392"] = &codec{id: 0xb351, name: "skein512-392", tag: "multihash"}
	reverseCodecs[0xb351] = codecs["skein512-392"]
	codecs["skein512-400"] = &codec{id: 0xb352, name: "skein512-400", tag: "multihash"}
	reverseCodecs[0xb352] = codecs["skein512-400"]
	codecs["skein512-408"] = &codec{id: 0xb353, name: "skein512-408", tag: "multihash"}
	reverseCodecs[0xb353] = codecs["skein512-408"]
	codecs["skein512-416"] = &codec{id: 0xb354, name: "skein512-416", tag: "multihash"}
	reverseCodecs[0xb354] = codecs["skein512-416"]
	codecs["skein512-424"] = &codec{id: 0xb355, name: "skein512-424", tag: "multihash"}
	reverseCodecs[0xb355] = codecs["skein512-424"]
	codecs["skein512-432"] = &codec{id: 0xb356, name: "skein512-432", tag: "multihash"}
	reverseCodecs[0xb356] = codecs["skein512-432"]
	codecs["skein512-440"] = &codec{id: 0xb357, name: "skein512-440", tag: "multihash"}
	reverseCodecs[0xb357] = codecs["skein512-440"]
	codecs["skein512-448"] = &codec{id: 0xb358, name: "skein512-448", tag: "multihash"}
	reverseCodecs[0xb358] = codecs["skein512-448"]
	codecs["skein512-456"] = &codec{id: 0xb359, name: "skein512-456", tag: "multihash"}
	reverseCodecs[0xb359] = codecs["skein512-456"]
	codecs["skein512-464"] = &codec{id: 0xb35a, name: "skein512-464", tag: "multihash"}
	reverseCodecs[0xb35a] = codecs["skein512-464"]
	codecs["skein512-472"] = &codec{id: 0xb35b, name: "skein512-472", tag: "multihash"}
	reverseCodecs[0xb35b] = codecs["skein512-472"]
	codecs["skein512-480"] = &codec{id: 0xb35c, name: "skein512-480", tag: "multihash"}
	reverseCodecs[0xb35c] = codecs["skein512-480"]
	codecs["skein512-488"] = &codec{id: 0xb35d, name: "skein512-488", tag: "multihash"}
	reverseCodecs[0xb35d] = codecs["skein512-488"]
	codecs["skein512-496"] = &codec{id: 0xb35e, name: "skein512-496", tag: "multihash"}
	reverseCodecs[0xb35e] = codecs["skein512-496"]
	codecs["skein512-504"] = &codec{id: 0xb35f, name: "skein512-504", tag: "multihash"}
	reverseCodecs[0xb35f] = codecs["skein512-504"]
	codecs["skein512-512"] = &codec{id: 0xb360, name: "skein512-512", tag: "multihash"}
	reverseCodecs[0xb360] = codecs["skein512-512"]
	codecs["skein1024-8"] = &codec{id: 0xb361, name: "skein1024-8", tag: "multihash"}
	reverseCodecs[0xb361] = codecs["skein1024-8"]
	codecs["skein1024-16"] = &codec{id: 0xb362, name: "skein1024-16", tag: "multihash"}
	reverseCodecs[0xb362] = codecs["skein1024-16"]
	codecs["skein1024-24"] = &codec{id: 0xb363, name: "skein1024-24", tag: "multihash"}
	reverseCodecs[0xb363] = codecs["skein1024-24"]
	codecs["skein1024-32"] = &codec{id: 0xb364, name: "skein1024-32", tag: "multihash"}
	reverseCodecs[0xb364] = codecs["skein1024-32"]
	codecs["skein1024-40"] = &codec{id: 0xb365, name: "skein1024-40", tag: "multihash"}
	reverseCodecs[0xb365] = codecs["skein1024-40"]
	codecs["skein1024-48"] = &codec{id: 0xb366, name: "skein1024-48", tag: "multihash"}
	reverseCodecs[0xb366] = codecs["skein1024-48"]
	codecs["skein1024-56"] = &codec{id: 0xb367, name: "skein1024-56", tag: "multihash"}
	reverseCodecs[0xb367] = codecs["skein1024-56"]
	codecs["skein1024-64"] = &codec{id: 0xb368, name: "skein1024-64", tag: "multihash"}
	reverseCodecs[0xb368] = codecs["skein1024-64"]
	codecs["skein1024-72"] = &codec{id: 0xb369, name: "skein1024-72", tag: "multihash"}
	reverseCodecs[0xb369] = codecs["skein1024-72"]
	codecs["skein1024-80"] = &codec{id: 0xb36a, name: "skein1024-80", tag: "multihash"}
	reverseCodecs[0xb36a] = codecs["skein1024-80"]
	codecs["skein1024-88"] = &codec{id: 0xb36b, name: "skein1024-88", tag: "multihash"}
	reverseCodecs[0xb36b] = codecs["skein1024-88"]
	codecs["skein1024-96"] = &codec{id: 0xb36c, name: "skein1024-96", tag: "multihash"}
	reverseCodecs[0xb36c] = codecs["skein1024-96"]
	codecs["skein1024-104"] = &codec{id: 0xb36d, name: "skein1024-104", tag: "multihash"}
	reverseCodecs[0xb36d] = codecs["skein1024-104"]
	codecs["skein1024-112"] = &codec{id: 0xb36e, name: "skein1024-112", tag: "multihash"}
	reverseCodecs[0xb36e] = codecs["skein1024-112"]
	codecs["skein1024-120"] = &codec{id: 0xb36f, name: "skein1024-120", tag: "multihash"}
	reverseCodecs[0xb36f] = codecs["skein1024-120"]
	codecs["skein1024-128"] = &codec{id: 0xb370, name: "skein1024-128", tag: "multihash"}
	reverseCodecs[0xb370] = codecs["skein1024-128"]
	codecs["skein1024-136"] = &codec{id: 0xb371, name: "skein1024-136", tag: "multihash"}
	reverseCodecs[0xb371] = codecs["skein1024-136"]
	codecs["skein1024-144"] = &codec{id: 0xb372, name: "skein1024-144", tag: "multihash"}
	reverseCodecs[0xb372] = codecs["skein1024-144"]
	codecs["skein1024-152"] = &codec{id: 0xb373, name: "skein1024-152", tag: "multihash"}
	reverseCodecs[0xb373] = codecs["skein1024-152"]
	codecs["skein1024-160"] = &codec{id: 0xb374, name: "skein1024-160", tag: "multihash"}
	reverseCodecs[0xb374] = codecs["skein1024-160"]
	codecs["skein1024-168"] = &codec{id: 0xb375, name: "skein1024-168", tag: "multihash"}
	reverseCodecs[0xb375] = codecs["skein1024-168"]
	codecs["skein1024-176"] = &codec{id: 0xb376, name: "skein1024-176", tag: "multihash"}
	reverseCodecs[0xb376] = codecs["skein1024-176"]
	codecs["skein1024-184"] = &codec{id: 0xb377, name: "skein1024-184", tag: "multihash"}
	reverseCodecs[0xb377] = codecs["skein1024-184"]
	codecs["skein1024-192"] = &codec{id: 0xb378, name: "skein1024-192", tag: "multihash"}
	reverseCodecs[0xb378] = codecs["skein1024-192"]
	codecs["skein1024-200"] = &codec{id: 0xb379, name: "skein1024-200", tag: "multihash"}
	reverseCodecs[0xb379] = codecs["skein1024-200"]
	codecs["skein1024-208"] = &codec{id: 0xb37a, name: "skein1024-208", tag: "multihash"}
	reverseCodecs[0xb37a] = codecs["skein1024-208"]
	codecs["skein1024-216"] = &codec{id: 0xb37b, name: "skein1024-216", tag: "multihash"}
	reverseCodecs[0xb37b] = codecs["skein1024-216"]
	codecs["skein1024-224"] = &codec{id: 0xb37c, name: "skein1024-224", tag: "multihash"}
	reverseCodecs[0xb37c] = codecs["skein1024-224"]
	codecs["skein1024-232"] = &codec{id: 0xb37d, name: "skein1024-232", tag: "multihash"}
	reverseCodecs[0xb37d] = codecs["skein1024-232"]
	codecs["skein1024-240"] = &codec{id: 0xb37e, name: "skein1024-240", tag: "multihash"}
	reverseCodecs[0xb37e] = codecs["skein1024-240"]
	codecs["skein1024-248"] = &codec{id: 0xb37f, name: "skein1024-248", tag: "multihash"}
	reverseCodecs[0xb37f] = codecs["skein1024-248"]
	codecs["skein1024-256"] = &codec{id: 0xb380, name: "skein1024-256", tag: "multihash"}
	reverseCodecs[0xb380] = codecs["skein1024-256"]
	codecs["skein1024-264"] = &codec{id: 0xb381, name: "skein1024-264", tag: "multihash"}
	reverseCodecs[0xb381] = codecs["skein1024-264"]
	codecs["skein1024-272"] = &codec{id: 0xb382, name: "skein1024-272", tag: "multihash"}
	reverseCodecs[0xb382] = codecs["skein1024-272"]
	codecs["skein1024-280"] = &codec{id: 0xb383, name: "skein1024-280", tag: "multihash"}
	reverseCodecs[0xb383] = codecs["skein1024-280"]
	codecs["skein1024-288"] = &codec{id: 0xb384, name: "skein1024-288", tag: "multihash"}
	reverseCodecs[0xb384] = codecs["skein1024-288"]
	codecs["skein1024-296"] = &codec{id: 0xb385, name: "skein1024-296", tag: "multihash"}
	reverseCodecs[0xb385] = codecs["skein1024-296"]
	codecs["skein1024-304"] = &codec{id: 0xb386, name: "skein1024-304", tag: "multihash"}
	reverseCodecs[0xb386] = codecs["skein1024-304"]
	codecs["skein1024-312"] = &codec{id: 0xb387, name: "skein1024-312", tag: "multihash"}
	reverseCodecs[0xb387] = codecs["skein1024-312"]
	codecs["skein1024-320"] = &codec{id: 0xb388, name: "skein1024-320", tag: "multihash"}
	reverseCodecs[0xb388] = codecs["skein1024-320"]
	codecs["skein1024-328"] = &codec{id: 0xb389, name: "skein1024-328", tag: "multihash"}
	reverseCodecs[0xb389] = codecs["skein1024-328"]
	codecs["skein1024-336"] = &codec{id: 0xb38a, name: "skein1024-336", tag: "multihash"}
	reverseCodecs[0xb38a] = codecs["skein1024-336"]
	codecs["skein1024-344"] = &codec{id: 0xb38b, name: "skein1024-344", tag: "multihash"}
	reverseCodecs[0xb38b] = codecs["skein1024-344"]
	codecs["skein1024-352"] = &codec{id: 0xb38c, name: "skein1024-352", tag: "multihash"}
	reverseCodecs[0xb38c] = codecs["skein1024-352"]
	codecs["skein1024-360"] = &codec{id: 0xb38d, name: "skein1024-360", tag: "multihash"}
	reverseCodecs[0xb38d] = codecs["skein1024-360"]
	codecs["skein1024-368"] = &codec{id: 0xb38e, name: "skein1024-368", tag: "multihash"}
	reverseCodecs[0xb38e] = codecs["skein1024-368"]
	codecs["skein1024-376"] = &codec{id: 0xb38f, name: "skein1024-376", tag: "multihash"}
	reverseCodecs[0xb38f] = codecs["skein1024-376"]
	codecs["skein1024-384"] = &codec{id: 0xb390, name: "skein1024-384", tag: "multihash"}
	reverseCodecs[0xb390] = codecs["skein1024-384"]
	codecs["skein1024-392"] = &codec{id: 0xb391, name: "skein1024-392", tag: "multihash"}
	reverseCodecs[0xb391] = codecs["skein1024-392"]
	codecs["skein1024-400"] = &codec{id: 0xb392, name: "skein1024-400", tag: "multihash"}
	reverseCodecs[0xb392] = codecs["skein1024-400"]
	codecs["skein1024-408"] = &codec{id: 0xb393, name: "skein1024-408", tag: "multihash"}
	reverseCodecs[0xb393] = codecs["skein1024-408"]
	codecs["skein1024-416"] = &codec{id: 0xb394, name: "skein1024-416", tag: "multihash"}
	reverseCodecs[0xb394] = codecs["skein1024-416"]
	codecs["skein1024-424"] = &codec{id: 0xb395, name: "skein1024-424", tag: "multihash"}
	reverseCodecs[0xb395] = codecs["skein1024-424"]
	codecs["skein1024-432"] = &codec{id: 0xb396, name: "skein1024-432", tag: "multihash"}
	reverseCodecs[0xb396] = codecs["skein1024-432"]
	codecs["skein1024-440"] = &codec{id: 0xb397, name: "skein1024-440", tag: "multihash"}
	reverseCodecs[0xb397] = codecs["skein1024-440"]
	codecs["skein1024-448"] = &codec{id: 0xb398, name: "skein1024-448", tag: "multihash"}
	reverseCodecs[0xb398] = codecs["skein1024-448"]
	codecs["skein1024-456"] = &codec{id: 0xb399, name: "skein1024-456", tag: "multihash"}
	reverseCodecs[0xb399] = codecs["skein1024-456"]
	codecs["skein1024-464"] = &codec{id: 0xb39a, name: "skein1024-464", tag: "multihash"}
	reverseCodecs[0xb39a] = codecs["skein1024-464"]
	codecs["skein1024-472"] = &codec{id: 0xb39b, name: "skein1024-472", tag: "multihash"}
	reverseCodecs[0xb39b] = codecs["skein1024-472"]
	codecs["skein1024-480"] = &codec{id: 0xb39c, name: "skein1024-480", tag: "multihash"}
	reverseCodecs[0xb39c] = codecs["skein1024-480"]
	codecs["skein1024-488"] = &codec{id: 0xb39d, name: "skein1024-488", tag: "multihash"}
	reverseCodecs[0xb39d] = codecs["skein1024-488"]
	codecs["skein1024-496"] = &codec{id: 0xb39e, name: "skein1024-496", tag: "multihash"}
	reverseCodecs[0xb39e] = codecs["skein1024-496"]
	codecs["skein1024-504"] = &codec{id: 0xb39f, name: "skein1024-504", tag: "multihash"}
	reverseCodecs[0xb39f] = codecs["skein1024-504"]
	codecs["skein1024-512"] = &codec{id: 0xb3a0, name: "skein1024-512", tag: "multihash"}
	reverseCodecs[0xb3a0] = codecs["skein1024-512"]
	codecs["skein1024-520"] = &codec{id: 0xb3a1, name: "skein1024-520", tag: "multihash"}
	reverseCodecs[0xb3a1] = codecs["skein1024-520"]
	codecs["skein1024-528"] = &codec{id: 0xb3a2, name: "skein1024-528", tag: "multihash"}
	reverseCodecs[0xb3a2] = codecs["skein1024-528"]
	codecs["skein1024-536"] = &codec{id: 0xb3a3, name: "skein1024-536", tag: "multihash"}
	reverseCodecs[0xb3a3] = codecs["skein1024-536"]
	codecs["skein1024-544"] = &codec{id: 0xb3a4, name: "skein1024-544", tag: "multihash"}
	reverseCodecs[0xb3a4] = codecs["skein1024-544"]
	codecs["skein1024-552"] = &codec{id: 0xb3a5, name: "skein1024-552", tag: "multihash"}
	reverseCodecs[0xb3a5] = codecs["skein1024-552"]
	codecs["skein1024-560"] = &codec{id: 0xb3a6, name: "skein1024-560", tag: "multihash"}
	reverseCodecs[0xb3a6] = codecs["skein1024-560"]
	codecs["skein1024-568"] = &codec{id: 0xb3a7, name: "skein1024-568", tag: "multihash"}
	reverseCodecs[0xb3a7] = codecs["skein1024-568"]
	codecs["skein1024-576"] = &codec{id: 0xb3a8, name: "skein1024-576", tag: "multihash"}
	reverseCodecs[0xb3a8] = codecs["skein1024-576"]
	codecs["skein1024-584"] = &codec{id: 0xb3a9, name: "skein1024-584", tag: "multihash"}
	reverseCodecs[0xb3a9] = codecs["skein1024-584"]
	codecs["skein1024-592"] = &codec{id: 0xb3aa, name: "skein1024-592", tag: "multihash"}
	reverseCodecs[0xb3aa] = codecs["skein1024-592"]
	codecs["skein1024-600"] = &codec{id: 0xb3ab, name: "skein1024-600", tag: "multihash"}
	reverseCodecs[0xb3ab] = codecs["skein1024-600"]
	codecs["skein1024-608"] = &codec{id: 0xb3ac, name: "skein1024-608", tag: "multihash"}
	reverseCodecs[0xb3ac] = codecs["skein1024-608"]
	codecs["skein1024-616"] = &codec{id: 0xb3ad, name: "skein1024-616", tag: "multihash"}
	reverseCodecs[0xb3ad] = codecs["skein1024-616"]
	codecs["skein1024-624"] = &codec{id: 0xb3ae, name: "skein1024-624", tag: "multihash"}
	reverseCodecs[0xb3ae] = codecs["skein1024-624"]
	codecs["skein1024-632"] = &codec{id: 0xb3af, name: "skein1024-632", tag: "multihash"}
	reverseCodecs[0xb3af] = codecs["skein1024-632"]
	codecs["skein1024-640"] = &codec{id: 0xb3b0, name: "skein1024-640", tag: "multihash"}
	reverseCodecs[0xb3b0] = codecs["skein1024-640"]
	codecs["skein1024-648"] = &codec{id: 0xb3b1, name: "skein1024-648", tag: "multihash"}
	reverseCodecs[0xb3b1] = codecs["skein1024-648"]
	codecs["skein1024-656"] = &codec{id: 0xb3b2, name: "skein1024-656", tag: "multihash"}
	reverseCodecs[0xb3b2] = codecs["skein1024-656"]
	codecs["skein1024-664"] = &codec{id: 0xb3b3, name: "skein1024-664", tag: "multihash"}
	reverseCodecs[0xb3b3] = codecs["skein1024-664"]
	codecs["skein1024-672"] = &codec{id: 0xb3b4, name: "skein1024-672", tag: "multihash"}
	reverseCodecs[0xb3b4] = codecs["skein1024-672"]
	codecs["skein1024-680"] = &codec{id: 0xb3b5, name: "skein1024-680", tag: "multihash"}
	reverseCodecs[0xb3b5] = codecs["skein1024-680"]
	codecs["skein1024-688"] = &codec{id: 0xb3b6, name: "skein1024-688", tag: "multihash"}
	reverseCodecs[0xb3b6] = codecs["skein1024-688"]
	codecs["skein1024-696"] = &codec{id: 0xb3b7, name: "skein1024-696", tag: "multihash"}
	reverseCodecs[0xb3b7] = codecs["skein1024-696"]
	codecs["skein1024-704"] = &codec{id: 0xb3b8, name: "skein1024-704", tag: "multihash"}
	reverseCodecs[0xb3b8] = codecs["skein1024-704"]
	codecs["skein1024-712"] = &codec{id: 0xb3b9, name: "skein1024-712", tag: "multihash"}
	reverseCodecs[0xb3b9] = codecs["skein1024-712"]
	codecs["skein1024-720"] = &codec{id: 0xb3ba, name: "skein1024-720", tag: "multihash"}
	reverseCodecs[0xb3ba] = codecs["skein1024-720"]
	codecs["skein1024-728"] = &codec{id: 0xb3bb, name: "skein1024-728", tag: "multihash"}
	reverseCodecs[0xb3bb] = codecs["skein1024-728"]
	codecs["skein1024-736"] = &codec{id: 0xb3bc, name: "skein1024-736", tag: "multihash"}
	reverseCodecs[0xb3bc] = codecs["skein1024-736"]
	codecs["skein1024-744"] = &codec{id: 0xb3bd, name: "skein1024-744", tag: "multihash"}
	reverseCodecs[0xb3bd] = codecs["skein1024-744"]
	codecs["skein1024-752"] = &codec{id: 0xb3be, name: "skein1024-752", tag: "multihash"}
	reverseCodecs[0xb3be] = codecs["skein1024-752"]
	codecs["skein1024-760"] = &codec{id: 0xb3bf, name: "skein1024-760", tag: "multihash"}
	reverseCodecs[0xb3bf] = codecs["skein1024-760"]
	codecs["skein1024-768"] = &codec{id: 0xb3c0, name: "skein1024-768", tag: "multihash"}
	reverseCodecs[0xb3c0] = codecs["skein1024-768"]
	codecs["skein1024-776"] = &codec{id: 0xb3c1, name: "skein1024-776", tag: "multihash"}
	reverseCodecs[0xb3c1] = codecs["skein1024-776"]
	codecs["skein1024-784"] = &codec{id: 0xb3c2, name: "skein1024-784", tag: "multihash"}
	reverseCodecs[0xb3c2] = codecs["skein1024-784"]
	codecs["skein1024-792"] = &codec{id: 0xb3c3, name: "skein1024-792", tag: "multihash"}
	reverseCodecs[0xb3c3] = codecs["skein1024-792"]
	codecs["skein1024-800"] = &codec{id: 0xb3c4, name: "skein1024-800", tag: "multihash"}
	reverseCodecs[0xb3c4] = codecs["skein1024-800"]
	codecs["skein1024-808"] = &codec{id: 0xb3c5, name: "skein1024-808", tag: "multihash"}
	reverseCodecs[0xb3c5] = codecs["skein1024-808"]
	codecs["skein1024-816"] = &codec{id: 0xb3c6, name: "skein1024-816", tag: "multihash"}
	reverseCodecs[0xb3c6] = codecs["skein1024-816"]
	codecs["skein1024-824"] = &codec{id: 0xb3c7, name: "skein1024-824", tag: "multihash"}
	reverseCodecs[0xb3c7] = codecs["skein1024-824"]
	codecs["skein1024-832"] = &codec{id: 0xb3c8, name: "skein1024-832", tag: "multihash"}
	reverseCodecs[0xb3c8] = codecs["skein1024-832"]
	codecs["skein1024-840"] = &codec{id: 0xb3c9, name: "skein1024-840", tag: "multihash"}
	reverseCodecs[0xb3c9] = codecs["skein1024-840"]
	codecs["skein1024-848"] = &codec{id: 0xb3ca, name: "skein1024-848", tag: "multihash"}
	reverseCodecs[0xb3ca] = codecs["skein1024-848"]
	codecs["skein1024-856"] = &codec{id: 0xb3cb, name: "skein1024-856", tag: "multihash"}
	reverseCodecs[0xb3cb] = codecs["skein1024-856"]
	codecs["skein1024-864"] = &codec{id: 0xb3cc, name: "skein1024-864", tag: "multihash"}
	reverseCodecs[0xb3cc] = codecs["skein1024-864"]
	codecs["skein1024-872"] = &codec{id: 0xb3cd, name: "skein1024-872", tag: "multihash"}
	reverseCodecs[0xb3cd] = codecs["skein1024-872"]
	codecs["skein1024-880"] = &codec{id: 0xb3ce, name: "skein1024-880", tag: "multihash"}
	reverseCodecs[0xb3ce] = codecs["skein1024-880"]
	codecs["skein1024-888"] = &codec{id: 0xb3cf, name: "skein1024-888", tag: "multihash"}
	reverseCodecs[0xb3cf] = codecs["skein1024-888"]
	codecs["skein1024-896"] = &codec{id: 0xb3d0, name: "skein1024-896", tag: "multihash"}
	reverseCodecs[0xb3d0] = codecs["skein1024-896"]
	codecs["skein1024-904"] = &codec{id: 0xb3d1, name: "skein1024-904", tag: "multihash"}
	reverseCodecs[0xb3d1] = codecs["skein1024-904"]
	codecs["skein1024-912"] = &codec{id: 0xb3d2, name: "skein1024-912", tag: "multihash"}
	reverseCodecs[0xb3d2] = codecs["skein1024-912"]
	codecs["skein1024-920"] = &codec{id: 0xb3d3, name: "skein1024-920", tag: "multihash"}
	reverseCodecs[0xb3d3] = codecs["skein1024-920"]
	codecs["skein1024-928"] = &codec{id: 0xb3d4, name: "skein1024-928", tag: "multihash"}
	reverseCodecs[0xb3d4] = codecs["skein1024-928"]
	codecs["skein1024-936"] = &codec{id: 0xb3d5, name: "skein1024-936", tag: "multihash"}
	reverseCodecs[0xb3d5] = codecs["skein1024-936"]
	codecs["skein1024-944"] = &codec{id: 0xb3d6, name: "skein1024-944", tag: "multihash"}
	reverseCodecs[0xb3d6] = codecs["skein1024-944"]
	codecs["skein1024-952"] = &codec{id: 0xb3d7, name: "skein1024-952", tag: "multihash"}
	reverseCodecs[0xb3d7] = codecs["skein1024-952"]
	codecs["skein1024-960"] = &codec{id: 0xb3d8, name: "skein1024-960", tag: "multihash"}
	reverseCodecs[0xb3d8] = codecs["skein1024-960"]
	codecs["skein1024-968"] = &codec{id: 0xb3d9, name: "skein1024-968", tag: "multihash"}
	reverseCodecs[0xb3d9] = codecs["skein1024-968"]
	codecs["skein1024-976"] = &codec{id: 0xb3da, name: "skein1024-976", tag: "multihash"}
	reverseCodecs[0xb3da] = codecs["skein1024-976"]
	codecs["skein1024-984"] = &codec{id: 0xb3db, name: "skein1024-984", tag: "multihash"}
	reverseCodecs[0xb3db] = codecs["skein1024-984"]
	codecs["skein1024-992"] = &codec{id: 0xb3dc, name: "skein1024-992", tag: "multihash"}
	reverseCodecs[0xb3dc] = codecs["skein1024-992"]
	codecs["skein1024-1000"] = &codec{id: 0xb3dd, name: "skein1024-1000", tag: "multihash"}
	reverseCodecs[0xb3dd] = codecs["skein1024-1000"]
	codecs["skein1024-1008"] = &codec{id: 0xb3de, name: "skein1024-1008", tag: "multihash"}
	reverseCodecs[0xb3de] = codecs["skein1024-1008"]
	codecs["skein1024-1016"] = &codec{id: 0xb3df, name: "skein1024-1016", tag: "multihash"}
	reverseCodecs[0xb3df] = codecs["skein1024-1016"]
	codecs["skein1024-1024"] = &codec{id: 0xb3e0, name: "skein1024-1024", tag: "multihash"}
	reverseCodecs[0xb3e0] = codecs["skein1024-1024"]
	codecs["poseidon-bls12_381-a2-fc1"] = &codec{id: 0xb401, name: "poseidon-bls12_381-a2-fc1", tag: "multihash"}
	reverseCodecs[0xb401] = codecs["poseidon-bls12_381-a2-fc1"]
	codecs["poseidon-bls12_381-a2-fc1-sc"] = &codec{id: 0xb402, name: "poseidon-bls12_381-a2-fc1-sc", tag: "multihash"}
	reverseCodecs[0xb402] = codecs["poseidon-bls12_381-a2-fc1-sc"]
	codecs["zeroxcert-imprint-256"] = &codec{id: 0xce11, name: "zeroxcert-imprint-256", tag: "zeroxcert"}
	reverseCodecs[0xce11] = codecs["zeroxcert-imprint-256"]
	codecs["fil-commitment-unsealed"] = &codec{id: 0xf101, name: "fil-commitment-unsealed", tag: "filecoin"}
	reverseCodecs[0xf101] = codecs["fil-commitment-unsealed"]
	codecs["fil-commitment-sealed"] = &codec{id: 0xf102, name: "fil-commitment-sealed", tag: "filecoin"}
	reverseCodecs[0xf102] = codecs["fil-commitment-sealed"]
	codecs["holochain-adr-v0"] = &codec{id: 0x807124, name: "holochain-adr-v0", tag: "holochain"}
	reverseCodecs[0x807124] = codecs["holochain-adr-v0"]
	codecs["holochain-adr-v1"] = &codec{id: 0x817124, name: "holochain-adr-v1", tag: "holochain"}
	reverseCodecs[0x817124] = codecs["holochain-adr-v1"]
	codecs["holochain-key-v0"] = &codec{id: 0x947124, name: "holochain-key-v0", tag: "holochain"}
	reverseCodecs[0x947124] = codecs["holochain-key-v0"]
	codecs["holochain-key-v1"] = &codec{id: 0x957124, name: "holochain-key-v1", tag: "holochain"}
	reverseCodecs[0x957124] = codecs["holochain-key-v1"]
	codecs["holochain-sig-v0"] = &codec{id: 0xa27124, name: "holochain-sig-v0", tag: "holochain"}
	reverseCodecs[0xa27124] = codecs["holochain-sig-v0"]
	codecs["holochain-sig-v1"] = &codec{id: 0xa37124, name: "holochain-sig-v1", tag: "holochain"}
	reverseCodecs[0xa37124] = codecs["holochain-sig-v1"]
	codecs["skynet-ns"] = &codec{id: 0xb19910, name: "skynet-ns", tag: "namespace"}
	reverseCodecs[0xb19910] = codecs["skynet-ns"]
}
